package gregapi.tileentity;

import static gregapi.data.CS.B;
import static gregapi.data.CS.F;
import static gregapi.data.CS.T;
import gregapi.GT_API;
import gregapi.code.ArrayListNoNulls;
import gregapi.util.UT;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import net.minecraft.block.Block;
import net.minecraft.init.Blocks;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.world.ChunkPosition;
import net.minecraft.world.World;

/**
 * @author Gregorius Techneticies
 */
public interface ITileEntityMachineBlockUpdateable {
	/** Called whenever a Machine Block adjacent to this Block got updated. */
	public void onMachineBlockUpdate();
	
	/** Utility for the Multi-Block-Updates */
	public static class Util {
		/** The List of Blocks, which can conduct Machine Block Updates */
	    public static final Map<Block, Integer> MACHINE_BLOCKS = new HashMap<Block, Integer>();
	    
		/**
		 * Causes a Machineblock Update
		 * This update will cause surrounding MultiBlock Machines to update their Configuration.
		 * You should call this Function in @Block.breakBlock and in @Block.onBlockAdded of your Machine.
		 * @param aWorld is being the World
		 * @param aX is the X-Coord of the update causing Block
		 * @param aY is the Y-Coord of the update causing Block
		 * @param aZ is the Z-Coord of the update causing Block
		 */
		public static boolean causeMachineUpdate(World aWorld, int aX, int aY, int aZ) {
			if (!aWorld.isRemote) new Thread(new MachineBlockUpdateRunnable(aWorld, aX, aY, aZ), "Machine Block Updating").start();
			return T;
		}
		
		/**
		 * Adds a Multi-Machine Block, like my Machine Casings for example.
		 * You should call @causeMachineUpdate in @Block.breakBlock and in @Block.onBlockAdded of your registered Block.
		 * You don't need to register TileEntities which implement @IMachineBlockUpdateable
		 * @param aID the ID of your Block
		 * @param aMeta the Metadata of the Blocks as Bitmask! -1 or ~0 for all Metavalues
		 */
		public static boolean registerMachineBlock(Block aBlock, int aMeta) {
			if (aBlock == null) return F;
	        if (GT_API.sCompatTC != null) GT_API.sCompatTC.registerPortholeBlacklistedBlock(aBlock);
			MACHINE_BLOCKS.put(aBlock, aMeta);
			return T;
		}
		
		/**
		 * Like above but with boolean Parameters instead of a BitMask
		 */
		public static boolean registerMachineBlock(Block aBlock, boolean... aMeta) {
			if (aBlock == null || aMeta == null || aMeta.length == 0) return F;
	        if (GT_API.sCompatTC != null) GT_API.sCompatTC.registerPortholeBlacklistedBlock(aBlock);
			int rMeta = 0;
			for (byte i = 0; i < aMeta.length && i < 16; i++) if (aMeta[i]) rMeta |= B[i];
			MACHINE_BLOCKS.put(aBlock, rMeta);
			return T;
		}
		
		/**
		 * if this Block is a Machine Update Conducting Block
		 */
		public static boolean isMachineBlock(Block aBlock, int aMeta) {
			if (aBlock == Blocks.air) return F;
			return (MACHINE_BLOCKS.containsKey(aBlock) && (MACHINE_BLOCKS.get(aBlock) & B[aMeta]) != 0);
		}
		
		private static class MachineBlockUpdateRunnable implements Runnable {
			private final int mX, mY, mZ;
			private final World mWorld;
			
			public MachineBlockUpdateRunnable(World aWorld, int aX, int aY, int aZ) {
				mWorld = aWorld; mX = aX; mY = aY; mZ = aZ;
			}
			
			@Override
			public void run() {
				try {stepToUpdateMachine(mWorld, mX, mY, mZ, new ArrayListNoNulls<ChunkPosition>());} catch(Throwable e) {/**/}
			}
			
			private static void stepToUpdateMachine(World aWorld, int aX, int aY, int aZ, ArrayList<ChunkPosition> aList) {
				aList.add(new ChunkPosition(aX, aY, aZ));
				TileEntity tTileEntity = UT.Worlds.getTileEntity(aWorld, aX, aY, aZ, true);
				if (tTileEntity != null && tTileEntity instanceof ITileEntityMachineBlockUpdateable) ((ITileEntityMachineBlockUpdateable)tTileEntity).onMachineBlockUpdate();
				if (aList.size() < 5 || (tTileEntity != null && tTileEntity instanceof ITileEntityMachineBlockUpdateable) || isMachineBlock(aWorld.getBlock(aX, aY, aZ), aWorld.getBlockMetadata(aX, aY, aZ))) {
					if (!aList.contains(new ChunkPosition(aX + 1, aY, aZ))) stepToUpdateMachine(aWorld, aX + 1, aY, aZ, aList);
					if (!aList.contains(new ChunkPosition(aX - 1, aY, aZ))) stepToUpdateMachine(aWorld, aX - 1, aY, aZ, aList);
					if (!aList.contains(new ChunkPosition(aX, aY + 1, aZ))) stepToUpdateMachine(aWorld, aX, aY + 1, aZ, aList);
					if (!aList.contains(new ChunkPosition(aX, aY - 1, aZ))) stepToUpdateMachine(aWorld, aX, aY - 1, aZ, aList);
					if (!aList.contains(new ChunkPosition(aX, aY, aZ + 1))) stepToUpdateMachine(aWorld, aX, aY, aZ + 1, aList);
					if (!aList.contains(new ChunkPosition(aX, aY, aZ - 1))) stepToUpdateMachine(aWorld, aX, aY, aZ - 1, aList);
				}
			}
		}
	}
}